use crate::prelude::*;

use anyhow::Result;

use cgmath::{Matrix, Matrix4, One};

use core::slice;
use std::{
    mem,
    sync::{Arc, Mutex},
};

enum AccelerationStructureBuilderData {
    TopLevel(Vec<VkAccelerationStructureInstanceKHR>),
    BottomLevel(Vec<VkAccelerationStructureGeometryKHR>, Vec<u32>),
}

pub struct AccelerationStructureBuilder {
    flags: VkAccelerationStructureCreateFlagBitsKHR,
    capture_replay_address: Option<VkDeviceAddress>,

    data: AccelerationStructureBuilderData,
}

impl AccelerationStructureBuilder {
    pub fn add_instance(
        mut self,
        blas: &Arc<AccelerationStructure>,
        transform: Option<Matrix4<f32>>,
        instance_flags: impl Into<VkGeometryInstanceFlagBitsKHR>,
    ) -> Self {
        match &mut self.data {
            AccelerationStructureBuilderData::TopLevel(instances) => {
                let transposed: [[f32; 4]; 4] = match transform {
                    Some(transform) => transform.transpose(),
                    None => Matrix4::one(),
                }
                .into();

                let instance = VkAccelerationStructureInstanceKHR::new(
                    VkTransformMatrixKHR::from(transposed),
                    instances.len() as u32,
                    0xFF,
                    0,
                    instance_flags,
                    blas.address(),
                );

                instances.push(instance);
            }
            AccelerationStructureBuilderData::BottomLevel(_, _) => {
                panic!("can not add acceleration structures to bottom level as")
            }
        }

        self
    }

    pub fn add_vertices<T: Send + Sync + 'static>(
        mut self,
        vertex_buffer: &Arc<Buffer<T>>,
        transform: Option<Arc<Buffer<Matrix4<f32>>>>,
        flags: impl Into<VkGeometryFlagBitsKHR>,
    ) -> Self {
        match &mut self.data {
            AccelerationStructureBuilderData::TopLevel(_) => {
                panic!("can not add buffers to top level as")
            }
            AccelerationStructureBuilderData::BottomLevel(geometries, primitive_counts) => {
                let geometry = VkAccelerationStructureGeometryKHR::new(
                    VK_GEOMETRY_TYPE_TRIANGLES_KHR,
                    VkAccelerationStructureGeometryDataKHR::from(
                        VkAccelerationStructureGeometryTrianglesDataKHR::new(
                            VK_FORMAT_R32G32B32_SFLOAT,
                            vertex_buffer.device_address().into(),
                            mem::size_of::<T>() as VkDeviceSize,
                            vertex_buffer.size() as u32,
                            VK_INDEX_TYPE_NONE_KHR,
                            VkDeviceOrHostAddressConstKHR::null(),
                            match &transform {
                                Some(transform_buffer) => transform_buffer.device_address().into(),
                                None => VkDeviceOrHostAddressConstKHR::null(),
                            },
                        ),
                    ),
                    flags,
                );

                let primitive_count = (vertex_buffer.size() / 3) as u32;

                geometries.push(geometry);
                primitive_counts.push(primitive_count);
            }
        }

        self
    }

    pub fn set_flags(mut self, flags: impl Into<VkAccelerationStructureCreateFlagBitsKHR>) -> Self {
        self.flags = flags.into();

        self
    }

    pub fn set_replay_address(mut self, capture_replay_address: VkDeviceAddress) -> Self {
        self.capture_replay_address = Some(capture_replay_address);

        self
    }

    pub fn build(
        self,
        device: Arc<Device>,
        recorder: &mut CommandBufferRecorder<'_>,
    ) -> Result<Arc<AccelerationStructure>> {
        let build_flags = VK_BUILD_ACCELERATION_STRUCTURE_PREFER_FAST_TRACE_BIT_KHR
            | VK_BUILD_ACCELERATION_STRUCTURE_ALLOW_UPDATE_BIT_KHR;

        let (acceleration_structure, generation_data, result_buffer, build_sizes_info) =
            match self.data {
                AccelerationStructureBuilderData::TopLevel(instances) => {
                    let instances_buffer = Buffer::builder()
                        .set_data(&instances)
                        .set_usage(
                            VK_BUFFER_USAGE_ACCELERATION_STRUCTURE_BUILD_INPUT_READ_ONLY_BIT_KHR
                                | VK_BUFFER_USAGE_SHADER_DEVICE_ADDRESS_BIT,
                        )
                        .set_memory_usage(MemoryUsage::CpuToGpu)
                        .build(device.clone())?;

                    let device_address: VkDeviceOrHostAddressConstKHR =
                        instances_buffer.device_address().into();

                    let geometry = VkAccelerationStructureGeometryKHR::new(
                        VK_GEOMETRY_TYPE_INSTANCES_KHR,
                        VkAccelerationStructureGeometryDataKHR::from(
                            VkAccelerationStructureGeometryInstancesDataKHR::from(device_address),
                        ),
                        VK_GEOMETRY_OPAQUE_BIT_KHR,
                    );

                    let mut geometry_info = VkAccelerationStructureBuildGeometryInfoKHR::minimal(
                        VK_ACCELERATION_STRUCTURE_TYPE_TOP_LEVEL_KHR,
                        build_flags,
                        VK_BUILD_ACCELERATION_STRUCTURE_MODE_BUILD_KHR,
                    );

                    geometry_info.set_geometry(slice::from_ref(&geometry));

                    let max_primitive_counts = 1;

                    let build_sizes_info = device.get_acceleration_structure_build_sizes(
                        VK_ACCELERATION_STRUCTURE_BUILD_TYPE_DEVICE_KHR,
                        &geometry_info,
                        &max_primitive_counts,
                    );

                    let result_buffer = Buffer::builder()
                        .set_usage(VK_BUFFER_USAGE_ACCELERATION_STRUCTURE_STORAGE_BIT_KHR)
                        .set_memory_usage(MemoryUsage::GpuOnly)
                        .set_size(build_sizes_info.accelerationStructureSize)
                        .build(device.clone())?;

                    let as_ci = VkAccelerationStructureCreateInfoKHR::new(
                        self.flags,
                        result_buffer.vk_handle(),
                        0,
                        build_sizes_info.accelerationStructureSize,
                        VK_ACCELERATION_STRUCTURE_TYPE_TOP_LEVEL_KHR,
                        self.capture_replay_address.unwrap_or(0),
                    );

                    (
                        device.create_acceleration_structure(&as_ci, None)?,
                        AccelerationStructureGenerationData::from((
                            instances,
                            geometry,
                            instances_buffer,
                        )),
                        result_buffer,
                        build_sizes_info,
                    )
                }
                AccelerationStructureBuilderData::BottomLevel(geometries, primitive_counts) => {
                    let mut geometry_info = VkAccelerationStructureBuildGeometryInfoKHR::minimal(
                        VK_ACCELERATION_STRUCTURE_TYPE_BOTTOM_LEVEL_KHR,
                        build_flags,
                        VK_BUILD_ACCELERATION_STRUCTURE_MODE_BUILD_KHR,
                    );

                    geometry_info.set_geometry(geometries.as_slice());

                    let max_primitive_counts = *primitive_counts
                        .iter()
                        .max()
                        .expect("empty primitive counts");

                    let build_sizes_info = device.get_acceleration_structure_build_sizes(
                        VK_ACCELERATION_STRUCTURE_BUILD_TYPE_DEVICE_KHR,
                        &geometry_info,
                        &max_primitive_counts,
                    );

                    let result_buffer = Buffer::builder()
                        .set_usage(VK_BUFFER_USAGE_ACCELERATION_STRUCTURE_STORAGE_BIT_KHR)
                        .set_memory_usage(MemoryUsage::GpuOnly)
                        .set_size(build_sizes_info.accelerationStructureSize)
                        .build(device.clone())?;

                    let as_ci = VkAccelerationStructureCreateInfoKHR::new(
                        self.flags,
                        result_buffer.vk_handle(),
                        0,
                        build_sizes_info.accelerationStructureSize,
                        VK_ACCELERATION_STRUCTURE_TYPE_BOTTOM_LEVEL_KHR,
                        self.capture_replay_address.unwrap_or(0),
                    );

                    (
                        device.create_acceleration_structure(&as_ci, None)?,
                        AccelerationStructureGenerationData::from((geometries, primitive_counts)),
                        result_buffer,
                        build_sizes_info,
                    )
                }
            };

        let acceleration_structure = Arc::new(AccelerationStructure {
            device: device.clone(),

            acceleration_structure,

            result_buffer,
            scratch_buffer: Mutex::new(AccelerationStructure::create_scratch_buffer(
                &device,
                build_sizes_info.buildScratchSize,
                device
                    .physical_device()
                    .acceleration_structure_properties()
                    .minAccelerationStructureScratchOffsetAlignment as VkDeviceSize,
            )?),

            update_scratch_buffer_size: build_sizes_info.updateScratchSize,

            generation_data,
            build_flags,
        });

        acceleration_structure.generate(
            recorder,
            VkAccelerationStructureKHR::NULL_HANDLE,
            VK_BUILD_ACCELERATION_STRUCTURE_MODE_BUILD_KHR,
        )?;

        Ok(acceleration_structure)
    }
}

pub struct AccelerationStructure {
    device: Arc<Device>,

    acceleration_structure: VkAccelerationStructureKHR,

    result_buffer: Arc<Buffer<u8>>,
    scratch_buffer: Mutex<Arc<Buffer<u8>>>,

    update_scratch_buffer_size: VkDeviceSize,

    generation_data: AccelerationStructureGenerationData,
    build_flags: VkBuildAccelerationStructureFlagBitsKHR,
}

impl AccelerationStructure {
    pub fn bottom_level() -> AccelerationStructureBuilder {
        AccelerationStructureBuilder {
            flags: 0.into(),
            capture_replay_address: None,

            data: AccelerationStructureBuilderData::BottomLevel(Vec::new(), Vec::new()),
        }
    }

    pub fn top_level() -> AccelerationStructureBuilder {
        AccelerationStructureBuilder {
            flags: 0.into(),
            capture_replay_address: None,

            data: AccelerationStructureBuilderData::TopLevel(Vec::new()),
        }
    }

    pub fn result_buffer(&self) -> &Arc<Buffer<u8>> {
        &self.result_buffer
    }

    fn address(&self) -> VkDeviceAddress {
        self.device.get_acceleration_structure_device_address(
            &VkAccelerationStructureDeviceAddressInfoKHR::new(self.acceleration_structure),
        )
    }

    pub fn update(&self, buffer_recorder: &mut CommandBufferRecorder<'_>) -> Result<()> {
        *self.scratch_buffer.lock().unwrap() = Self::create_scratch_buffer(
            &self.device,
            self.update_scratch_buffer_size,
            self.device
                .physical_device()
                .acceleration_structure_properties()
                .minAccelerationStructureScratchOffsetAlignment as VkDeviceSize,
        )?;

        self.generate(
            buffer_recorder,
            self.acceleration_structure,
            VK_BUILD_ACCELERATION_STRUCTURE_MODE_UPDATE_KHR,
        )
    }

    fn generate(
        &self,
        buffer_recorder: &mut CommandBufferRecorder<'_>,
        src: VkAccelerationStructureKHR,
        mode: VkBuildAccelerationStructureModeKHR,
    ) -> Result<()> {
        match &self.generation_data {
            AccelerationStructureGenerationData::TopLevel(instances, geometry, _buffer) => {
                let mut build_info = VkAccelerationStructureBuildGeometryInfoKHR::new(
                    VK_ACCELERATION_STRUCTURE_TYPE_TOP_LEVEL_KHR,
                    self.build_flags,
                    mode,
                    src,
                    self.acceleration_structure,
                    self.scratch_buffer.lock().unwrap().device_address().into(),
                );

                let geometry_slice: &[&VkAccelerationStructureGeometryKHR] = &[geometry];

                build_info.set_geometry(geometry_slice);

                buffer_recorder.build_acceleration_structures(
                    &[build_info],
                    &[&[VkAccelerationStructureBuildRangeInfoKHR::new(
                        instances.len() as u32,
                        0,
                        0,
                        0,
                    )]],
                );
            }
            AccelerationStructureGenerationData::BottomLevel(geometries, range_infos) => {
                let mut build_info = VkAccelerationStructureBuildGeometryInfoKHR::new(
                    VK_ACCELERATION_STRUCTURE_TYPE_BOTTOM_LEVEL_KHR,
                    self.build_flags,
                    mode,
                    src,
                    self.acceleration_structure,
                    self.scratch_buffer.lock().unwrap().device_address().into(),
                );

                build_info.set_geometry(geometries.as_slice());

                buffer_recorder
                    .build_acceleration_structures(&[build_info], &[range_infos.as_slice()]);
            }
        };

        buffer_recorder.memory_barrier(
            VK_ACCESS_ACCELERATION_STRUCTURE_WRITE_BIT_KHR,
            VK_PIPELINE_STAGE_ACCELERATION_STRUCTURE_BUILD_BIT_KHR,
            VK_ACCESS_ACCELERATION_STRUCTURE_WRITE_BIT_KHR,
            VK_PIPELINE_STAGE_ACCELERATION_STRUCTURE_BUILD_BIT_KHR,
        );

        Ok(())
    }

    #[inline]
    fn create_scratch_buffer(
        device: &Arc<Device>,
        size: VkDeviceSize,
        alignment: VkDeviceSize,
    ) -> Result<Arc<Buffer<u8>>> {
        Buffer::builder()
            .set_usage(
                VK_BUFFER_USAGE_STORAGE_BUFFER_BIT | VK_BUFFER_USAGE_SHADER_DEVICE_ADDRESS_BIT,
            )
            .set_memory_usage(MemoryUsage::GpuOnly)
            .set_size(size)
            .force_alignment(alignment)
            .build(device.clone())
    }
}

impl Drop for AccelerationStructure {
    fn drop(&mut self) {
        self.device
            .destroy_acceleration_structure(self.acceleration_structure, None);
    }
}

impl_vk_handle!(
    AccelerationStructure,
    VkAccelerationStructureKHR,
    acceleration_structure
);

enum AccelerationStructureGenerationData {
    TopLevel(
        Vec<VkAccelerationStructureInstanceKHR>,
        VkAccelerationStructureGeometryKHR,
        Arc<Buffer<VkAccelerationStructureInstanceKHR>>,
    ),
    BottomLevel(
        Vec<VkAccelerationStructureGeometryKHR>,
        Vec<VkAccelerationStructureBuildRangeInfoKHR>,
    ),
}

impl
    From<(
        Vec<VkAccelerationStructureInstanceKHR>,
        VkAccelerationStructureGeometryKHR,
        Arc<Buffer<VkAccelerationStructureInstanceKHR>>,
    )> for AccelerationStructureGenerationData
{
    fn from(
        (instances, geometry, buffer): (
            Vec<VkAccelerationStructureInstanceKHR>,
            VkAccelerationStructureGeometryKHR,
            Arc<Buffer<VkAccelerationStructureInstanceKHR>>,
        ),
    ) -> Self {
        Self::TopLevel(instances, geometry, buffer)
    }
}

impl From<(Vec<VkAccelerationStructureGeometryKHR>, Vec<u32>)>
    for AccelerationStructureGenerationData
{
    fn from(
        (geometries, primitive_counts): (Vec<VkAccelerationStructureGeometryKHR>, Vec<u32>),
    ) -> Self {
        Self::BottomLevel(
            geometries,
            primitive_counts
                .iter()
                .map(|&count| VkAccelerationStructureBuildRangeInfoKHR::new(count, 0, 0, 0))
                .collect(),
        )
    }
}
