use crate::prelude::*;
use anyhow::Result;
use std::sync::{
    atomic::{AtomicU32, AtomicUsize, Ordering::SeqCst},
    Arc,
};

pub mod sub_pass;
use sub_pass::{AttachmentInfo, AttachmentInfoUsage, SubPass};

pub struct RenderTargetBuilder {
    sub_passes: Vec<SubPass>,
}

impl RenderTargetBuilder {
    pub fn add_sub_pass(mut self, sub_pass: SubPass) -> Self {
        self.sub_passes.push(sub_pass);

        self
    }

    pub fn build(self, device: &Arc<Device>) -> Result<RenderTarget> {
        #[cfg(debug_assertions)]
        {
            // sub passes must not be empty
            assert!(!self.sub_passes.is_empty());

            // sub passes must all have the same extent
            let first_extent = self.sub_passes[0].extent();

            for sub_pass in self.sub_passes.iter() {
                assert!(sub_pass.extent() == first_extent);
            }
        }

        // create render pass

        // gather attachment descriptions
        let mut attachments = Vec::new();

        self.map_attachment(|attachment| {
            attachments.push(attachment.description.clone());
        });

        // create attachment references
        let mut attachment_references: Vec<SubPassAttachmentReferences> = Vec::new();
        let mut attachment_index = 0;

        // gather all color, depth and resolve attachment and add input attachments from previous sup passes
        for sub_pass in self.sub_passes.iter() {
            let mut references = SubPassAttachmentReferences::default();
            references.offset = attachment_index as usize;

            for attachment in sub_pass.attachments().iter() {
                let attachment_reference = VkAttachmentReference {
                    attachment: attachment_index,
                    layout: attachment.layout,
                };

                match attachment.usage {
                    AttachmentInfoUsage::Output => {
                        references.color_attachments.push(attachment_reference);
                    }
                    AttachmentInfoUsage::Resolve => {
                        references.resolve_attachments.push(attachment_reference);
                    }
                    AttachmentInfoUsage::Depth => {
                        // make sure only 1 depth attachment is used per subpass
                        debug_assert!(
                            references.depth_stencil_attachment.is_none(),
                            "only 1 depth attachment per sub pass allowed"
                        );

                        references.depth_stencil_attachment = Some(attachment_reference);
                    }
                }

                attachment_index += 1;
            }

            // check if input infos are set
            if let Some(input_info) = sub_pass.inputs() {
                debug_assert!(input_info.sub_pass_index < attachment_references.len());

                let input_pass_references = &attachment_references[input_info.sub_pass_index];

                for input_index in input_info.input_indices.iter() {
                    references.input_attachments.push(VkAttachmentReference {
                        attachment: (input_index + input_pass_references.offset) as u32,
                        layout: VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL,
                    });
                }
            }

            attachment_references.push(references);
        }

        // gather sub pass descriptions
        let mut descriptions = Vec::new();

        for attachment_reference in attachment_references.iter() {
            descriptions.push(VkSubpassDescription::new(
                0,
                &attachment_reference.input_attachments,
                &attachment_reference.color_attachments,
                &attachment_reference.resolve_attachments,
                attachment_reference.depth_stencil_attachment.as_ref(),
                &[],
            ));
        }

        // gather sub pass dependencies
        let mut dependencies = Vec::new();

        dependencies.push(VkSubpassDependency::new(
            VK_SUBPASS_EXTERNAL,
            0,
            VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT,
            VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT,
            VK_ACCESS_MEMORY_READ_BIT,
            VK_ACCESS_COLOR_ATTACHMENT_READ_BIT | VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT,
            VK_DEPENDENCY_BY_REGION_BIT,
        ));

        for (index, sub_pass) in self.sub_passes.iter().enumerate() {
            dependencies.push(VkSubpassDependency::new(
                index as u32,
                index as u32 + 1,
                VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT,
                CommandBuffer::access_to_stage(sub_pass.output_usage()),
                VK_ACCESS_COLOR_ATTACHMENT_READ_BIT | VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT,
                sub_pass.output_usage(),
                VK_DEPENDENCY_BY_REGION_BIT,
            ));
        }

        if let Some(last_dependency) = dependencies.last_mut() {
            last_dependency.dstSubpass = VK_SUBPASS_EXTERNAL;
            last_dependency.dstStageMask = VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT.into();
            last_dependency.dstAccessMask = VK_ACCESS_MEMORY_READ_BIT.into();
        }

        // dependencies.push(VkSubpassDependency::new(
        //     self.sub_passes.len() as u32 - 1,
        //     VK_SUBPASS_EXTERNAL,
        //     VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT,
        //     VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT,
        //     VK_ACCESS_COLOR_ATTACHMENT_READ_BIT | VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT,
        //     VK_ACCESS_MEMORY_READ_BIT,
        //     VK_DEPENDENCY_BY_REGION_BIT,
        // ));

        let render_pass =
            RenderPass::new(device.clone(), &descriptions, &attachments, &dependencies)?;

        // create frame buffers
        let max_images = self.max_images();
        let extent = self.sub_passes[0].extent();

        let framebuffers: Result<Vec<Arc<Framebuffer>>> = (0..max_images)
            .map(|i| {
                let mut framebuffer_builder = Framebuffer::builder()
                    .set_render_pass(&render_pass)
                    .set_width(extent.width)
                    .set_height(extent.height);

                for sub_pass in self.sub_passes.iter() {
                    for attachment in sub_pass.attachments().iter() {
                        framebuffer_builder =
                            framebuffer_builder.add_attachment(attachment.image(i));
                    }
                }

                framebuffer_builder.build(render_pass.device().clone())
            })
            .collect();

        let mut clear_values = Vec::new();

        self.map_attachment(|attachment| {
            clear_values.push(attachment.clear_value.clone());
        });

        Ok(RenderTarget {
            render_pass,
            framebuffers: framebuffers?,
            clear_values,

            extent,

            sub_passes: self.sub_passes,

            current_subpass: AtomicU32::new(0),
            framebuffer_index: AtomicUsize::new(0),
        })
    }

    #[inline]
    fn max_images(&self) -> usize {
        let mut max_images = 0;

        for sub_pass in self.sub_passes.iter() {
            max_images = max_images.max(sub_pass.max_images_per_attachment());
        }

        max_images
    }

    #[inline]
    fn map_attachment<'a, F>(&'a self, mut f: F)
    where
        F: FnMut(&'a AttachmentInfo) -> (),
    {
        for sub_pass in self.sub_passes.iter() {
            for attachment in sub_pass.attachments().iter() {
                f(attachment);
            }
        }
    }
}

#[derive(Default)]
struct SubPassAttachmentReferences {
    offset: usize,

    input_attachments: Vec<VkAttachmentReference>,
    color_attachments: Vec<VkAttachmentReference>,

    resolve_attachments: Vec<VkAttachmentReference>,
    depth_stencil_attachment: Option<VkAttachmentReference>,
}

#[derive(Debug)]
pub struct RenderTarget {
    render_pass: Arc<RenderPass>,
    framebuffers: Vec<Arc<Framebuffer>>,
    clear_values: Vec<VkClearValue>,

    extent: VkExtent2D,

    sub_passes: Vec<SubPass>,

    current_subpass: AtomicU32,
    framebuffer_index: AtomicUsize,
}

impl RenderTarget {
    pub fn builder() -> RenderTargetBuilder {
        RenderTargetBuilder {
            sub_passes: Vec::new(),
        }
    }

    pub fn render_pass(&self) -> &Arc<RenderPass> {
        &self.render_pass
    }

    pub fn framebuffer(&self, index: usize) -> &Arc<Framebuffer> {
        &self.framebuffers[index]
    }

    pub fn sub_pass(&self, index: usize) -> &SubPass {
        &self.sub_passes[index]
    }

    pub fn width(&self) -> u32 {
        self.extent.width
    }

    pub fn height(&self) -> u32 {
        self.extent.height
    }

    pub(crate) fn subpass_index(&self) -> u32 {
        self.current_subpass.load(SeqCst)
    }

    pub fn inheritance_info(&self) -> VkCommandBufferInheritanceInfo {
        CommandBuffer::inheritance_info(
            Some(&self.render_pass),
            Some(self.current_subpass.load(SeqCst)),
            Some(self.framebuffer(self.framebuffer_index.load(SeqCst))),
            None,
        )
    }

    pub fn begin<'a>(
        &'a self,
        buffer_recorder: &'a mut CommandBufferRecorder<'a>,
        subpass_content: VkSubpassContents,
        framebuffer_index: usize,
    ) {
        let renderpass_begin = VkRenderPassBeginInfo::new(
            self.render_pass.vk_handle(),
            self.framebuffers[framebuffer_index].vk_handle(),
            VkRect2D {
                offset: VkOffset2D { x: 0, y: 0 },
                extent: self.extent,
            },
            self.clear_values.as_slice(),
        );

        buffer_recorder.begin_render_pass(renderpass_begin, subpass_content);
        buffer_recorder.set_render_target(self);

        self.framebuffer_index.store(framebuffer_index, SeqCst);
        self.current_subpass.store(0, SeqCst);
    }

    pub fn next_subpass(
        &self,
        buffer_recorder: &CommandBufferRecorder<'_>,
        subpass_content: VkSubpassContents,
    ) {
        buffer_recorder.next_subpass(subpass_content);
        self.current_subpass.fetch_add(1, SeqCst);
    }

    pub fn end<'a>(&self, buffer_recorder: &'a mut CommandBufferRecorder<'a>) {
        buffer_recorder.end_render_pass();
        buffer_recorder.clear_render_target();
    }
}

// impl<'a> RenderTargetBuilder<'a> {
//     fn create_images_and_renderpass(&self, device: &Arc<Device>) -> Result<Arc<RenderPass>> {
//         let subpass_descriptions = [match resolve_reference {
//             Some(resvole_ref) => VkSubpassDescription::new(
//                 0,
//                 &[],
//                 color_references.as_slice(),
//                 &[resvole_ref],
//                 match depth_reference {
//                     Some(ref depth_ref) => Some(depth_ref),
//                     None => None,
//                 },
//                 &[],
//             ),
//             None => VkSubpassDescription::new(
//                 0,
//                 &[],
//                 color_references.as_slice(),
//                 &[],
//                 match depth_reference {
//                     Some(ref depth_ref) => Some(depth_ref),
//                     None => None,
//                 },
//                 &[],
//             ),
//         }];

//         let dependencies = if color_references.is_empty() {
//             // assume, that when no color references are given,
//             // we want to store the depth information for later
//             if depth_reference.is_some() {
//                 for attachment in &mut attachments {
//                     if attachment.format == VK_FORMAT_D16_UNORM {
//                         attachment.storeOp = VK_ATTACHMENT_STORE_OP_STORE;
//                         attachment.finalLayout = VK_IMAGE_LAYOUT_DEPTH_STENCIL_READ_ONLY_OPTIMAL;
//                         break;
//                     }
//                 }
//             }

//             [
//                 VkSubpassDependency::new(
//                     VK_SUBPASS_EXTERNAL,
//                     0,
//                     VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT,
//                     VK_PIPELINE_STAGE_LATE_FRAGMENT_TESTS_BIT,
//                     0,
//                     VK_ACCESS_DEPTH_STENCIL_ATTACHMENT_READ_BIT
//                         | VK_ACCESS_DEPTH_STENCIL_ATTACHMENT_WRITE_BIT,
//                     VK_DEPENDENCY_BY_REGION_BIT,
//                 ),
//                 VkSubpassDependency::new(
//                     0,
//                     VK_SUBPASS_EXTERNAL,
//                     VK_PIPELINE_STAGE_LATE_FRAGMENT_TESTS_BIT,
//                     VK_PIPELINE_STAGE_FRAGMENT_SHADER_BIT,
//                     VK_ACCESS_DEPTH_STENCIL_ATTACHMENT_READ_BIT
//                         | VK_ACCESS_DEPTH_STENCIL_ATTACHMENT_WRITE_BIT,
//                     VK_ACCESS_SHADER_READ_BIT,
//                     VK_DEPENDENCY_BY_REGION_BIT,
//                 ),
//             ]
//         } else {
//             [
//                 VkSubpassDependency::new(
//                     VK_SUBPASS_EXTERNAL,
//                     0,
//                     VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT,
//                     VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT,
//                     VK_ACCESS_MEMORY_READ_BIT,
//                     VK_ACCESS_COLOR_ATTACHMENT_READ_BIT | VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT,
//                     VK_DEPENDENCY_BY_REGION_BIT,
//                 ),
//                 VkSubpassDependency::new(
//                     0,
//                     VK_SUBPASS_EXTERNAL,
//                     VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT,
//                     VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT,
//                     VK_ACCESS_COLOR_ATTACHMENT_READ_BIT | VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT,
//                     VK_ACCESS_MEMORY_READ_BIT,
//                     VK_DEPENDENCY_BY_REGION_BIT,
//                 ),
//             ]
//         };

//         let renderpass = RenderPass::new(
//             device.clone(),
//             &subpass_descriptions,
//             attachments.as_slice(),
//             &dependencies,
//         )?;

//         Ok(renderpass)
//     }
// }
